import api from "@/services/core/api";
import { composeImgUrl } from "@/utils";
import { Canvas, CanvasProps } from "@tarojs/components"
import Taro, { CanvasContext } from "@tarojs/taro";
import { forwardRef, useEffect, useImperativeHandle, useRef } from "react";

type ElectronicSignatureProps = {
    width: string
    height: string
    lineColor?: string
    lineSize?: number  // 粗细
    bgColor?: string
    defaultSign?: string
}

export type ElectronicSignatureRef = {
    subCanvas: () => Promise<any>
    previewSign: () => void
    resetDraw: () => void
}

const canvasName = 'mySignCanvas'
const transparent = 1  // 透明度
const lineMin = 0.5   // 最小笔画半径
const lineMax = 4   // 最大笔画半径
const pressure = 1   // 默认压力
const smoothness = 60  // 顺滑度，用60的距离来计算速度
const ElectronicSignature = forwardRef<ElectronicSignatureRef, ElectronicSignatureProps>((props, ref) => {
    const { width, height, lineColor = '#1A1A1A', lineSize = 1.5, bgColor = '#fff', defaultSign } = props


    const canvasWidth = useRef(0)
    const canvasHeight = useRef(0)
    const currentPoint = useRef<any>({})
    const currentLine = useRef<any[]>([])  // 当前线条
    const firstTouch = useRef<boolean>(true)
    const radius = useRef(1) //画圆的半径

    const cutArea = useRef({ top: 0, right: 0, bottom: 0, left: 0 }) //裁剪区域
    const lastPoint = useRef([])
    const chirography = useRef<any[]>([])  // 笔迹
    const linePrack = useRef<any[]>([])  // 划线轨迹 , 生成线条的实际点

    const ctxx = useRef<CanvasContext>()
    useEffect(() => {
        Taro.nextTick(() => {
            ctxx.current = Taro.createCanvasContext(canvasName);
            Taro.createSelectorQuery().select(`.${canvasName}`).boundingClientRect((rect) => {
                // @ts-ignore
                canvasWidth.current = rect.width || rect[0].width
                // @ts-ignore
                canvasHeight.current = rect.height || rect[0].height
            }).exec()
        })
    }, [])

    const isFirst = useRef(1)
    useEffect(() => {
        if (isFirst.current === 1) {
            isFirst.current += 1
            setTimeout(() => {
                if (defaultSign) {
                    Taro.downloadFile({
                        url: api.assetsUrl + defaultSign,
                        success: (res) => {
                            ctxx.current?.drawImage(res.tempFilePath, 0, 0, canvasWidth.current, canvasHeight.current)
                            ctxx.current?.draw()
                        }
                    })
                } else {
                    resetDraw()
                }
            }, 200)
        }
    }, [defaultSign])

    // 手指触摸动作开始
    const uploadScaleStart: CanvasProps['onTouchStart'] = (e) => {
        if (e.type != 'touchstart') return false;
        let ctx = ctxx.current!
        ctx.setFillStyle(lineColor); // 初始线条设置颜色
        ctx.setGlobalAlpha(transparent); // 设置半透明
        let nowCurrentPoint = {
            x: e.touches[0].x,
            y: e.touches[0].y
        };
        currentLine.current.unshift({
            time: new Date().getTime(),
            dis: 0,
            x: nowCurrentPoint.x,
            y: nowCurrentPoint.y
        });
        currentPoint.current = nowCurrentPoint;
        // currentLine
        if (firstTouch.current) {
            cutArea.current = {
                top: nowCurrentPoint.y,
                right: nowCurrentPoint.x,
                bottom: nowCurrentPoint.y,
                left: nowCurrentPoint.x
            };
            firstTouch.current = false;
        }
        pointToLine(currentLine.current);
    }

    // 手指触摸后移动
    const uploadScaleMove: CanvasProps['onTouchMove'] = (e) => {
        if (e.type != 'touchmove') return false;
        // if (e.cancelable) {
        //     // 判断默认行为是否已经被禁用
        //     if (!e.defaultPrevented) {
        //         e.preventDefault();
        //     }
        // }
        let point = {
            x: e.touches[0].x,
            y: e.touches[0].y
        };

        //测试裁剪
        if (point.y < cutArea.current.top) {
            cutArea.current.top = point.y;
        }
        if (point.y < 0) cutArea.current.top = 0;

        if (point.x > cutArea.current.right) {
            cutArea.current.right = point.x;
        }
        if (canvasWidth.current - point.x <= 0) {
            cutArea.current.right = canvasWidth.current;
        }
        if (point.y > cutArea.current.bottom) {
            cutArea.current.bottom = point.y;
        }
        if (canvasHeight.current - point.y <= 0) {
            cutArea.current.bottom = canvasHeight.current;
        }
        if (point.x < cutArea.current.left) {
            cutArea.current.left = point.x;
        }
        if (point.x < 0) cutArea.current.left = 0;

        lastPoint.current = currentPoint.current
        currentPoint.current = point

        currentLine.current.unshift({
            time: new Date().getTime(),
            dis: distance(currentPoint.current, lastPoint.current),
            x: point.x,
            y: point.y
        });

        pointToLine(currentLine.current);
    }

    // 手指触摸动作结束
    const uploadScaleEnd: CanvasProps['onTouchEnd'] = (e) => {
        if (e.type != 'touchend') return 0;
        let point = {
            x: e.changedTouches[0].x,
            y: e.changedTouches[0].y
        };
        lastPoint.current = currentPoint.current
        currentPoint.current = point

        currentLine.current.unshift({
            time: new Date().getTime(),
            dis: distance(currentPoint.current, lastPoint.current),
            x: point.x,
            y: point.y
        });

        if (currentLine.current.length > 2) {
            // const info = (currentLine[0].time - currentLine[currentLine.current.length - 1].time) / currentLine.current.length;
            //$("#info").text(info.toFixed(2));
        }
        //一笔结束，保存笔迹的坐标点，清空，当前笔迹
        //增加判断是否在手写区域；
        pointToLine(currentLine.current);
        chirography.current.unshift({ lineSize, lineColor });

        linePrack.current.unshift(currentLine.current);
        currentLine.current = [];
    }

    //画两点之间的线条；参数为:line，会绘制最近的开始的两个点；
    const pointToLine = (line: any[]) => {
        calcBethelLine(line);
        return;
    }
    //计算插值的方式；
    const calcBethelLine = (line: any[]) => {
        if (line.length <= 1) {
            line[0].r = radius;
            return;
        }
        let x0,
            x1,
            x2,
            y0,
            y1,
            y2,
            r0,
            r1,
            r2,
            len,
            lastRadius = radius.current,
            dis = 0,
            time = 0,
            curveValue = 0.5;
        if (line.length <= 2) {
            x0 = line[1].x;
            y0 = line[1].y;
            x2 = line[1].x + (line[0].x - line[1].x) * curveValue;
            y2 = line[1].y + (line[0].y - line[1].y) * curveValue;
            //x2 = line[1].x;
            //y2 = line[1].y;
            x1 = x0 + (x2 - x0) * curveValue;
            y1 = y0 + (y2 - y0) * curveValue;
        } else {
            x0 = line[2].x + (line[1].x - line[2].x) * curveValue;
            y0 = line[2].y + (line[1].y - line[2].y) * curveValue;
            x1 = line[1].x;
            y1 = line[1].y;
            x2 = x1 + (line[0].x - x1) * curveValue;
            y2 = y1 + (line[0].y - y1) * curveValue;
        }
        //从计算公式看，三个点分别是(x0,y0),(x1,y1),(x2,y2) ；(x1,y1)这个是控制点，控制点不会落在曲线上；实际上，这个点还会手写获取的实际点，却落在曲线上
        len = distance({
            x: x2,
            y: y2
        }, {
            x: x0,
            y: y0
        });
        for (let n = 0; n < line.length - 1; n++) {
            dis += line[n].dis;
            time += line[n].time - line[n + 1].time;
            if (dis > smoothness) break;
        }

        radius.current = Math.min((time / len) * pressure + lineMin, lineMax) * lineSize;
        line[0].r = radius.current
        //计算笔迹半径；
        if (line.length <= 2) {
            r0 = (lastRadius + radius.current) / 2;
            r1 = r0;
            r2 = r1;
            //return;
        } else {
            r0 = (line[2].r + line[1].r) / 2;
            r1 = line[1].r;
            r2 = (line[1].r + line[0].r) / 2;
        }
        let n = 5;
        let point: { x: number, y: number, r: number }[] = [];
        for (let i = 0; i < n; i++) {
            let t = i / (n - 1);
            let x = (1 - t) * (1 - t) * x0 + 2 * t * (1 - t) * x1 + t * t * x2;
            let y = (1 - t) * (1 - t) * y0 + 2 * t * (1 - t) * y1 + t * t * y2;
            let r = lastRadius + ((radius.current - lastRadius) / n) * i;
            point.push({ x, y, r });
            if (point.length === 3) {
                let a = ctaCalc(point[0].x, point[0].y, point[0].r, point[1].x, point[1].y, point[1].r, point[2].x, point[2].y, point[2].r)
                a[0].color = lineColor
                // let bethelPoint = this.bethelPoint;
                // bethelPoint = bethelPoint.push(a);
                bethelDraw(a, 1);
                point = [{ x, y, r }]
            }
        }
        currentLine.current = line
    }

    //求两点之间距离
    const distance = (a, b) => {
        let x = b.x - a.x;
        let y = b.y - a.y;
        return Math.sqrt(x * x + y * y);
    }
    const ctaCalc = (x0, y0, r0, x1, y1, r1, x2, y2, r2) => {
        let a: any[] = [],
            vx01,
            vy01,
            norm,
            n_x0,
            n_y0,
            vx21,
            vy21,
            n_x2,
            n_y2;
        vx01 = x1 - x0;
        vy01 = y1 - y0;
        norm = Math.sqrt(vx01 * vx01 + vy01 * vy01 + 0.0001) * 2;
        vx01 = (vx01 / norm) * r0;
        vy01 = (vy01 / norm) * r0;
        n_x0 = vy01;
        n_y0 = -vx01;
        vx21 = x1 - x2;
        vy21 = y1 - y2;
        norm = Math.sqrt(vx21 * vx21 + vy21 * vy21 + 0.0001) * 2;
        vx21 = (vx21 / norm) * r2;
        vy21 = (vy21 / norm) * r2;
        n_x2 = -vy21;
        n_y2 = vx21;
        a.push({
            mx: x0 + n_x0,
            my: y0 + n_y0,
            color: '#1A1A1A'
        });
        a.push({
            c1x: x1 + n_x0,
            c1y: y1 + n_y0,
            c2x: x1 + n_x2,
            c2y: y1 + n_y2,
            ex: x2 + n_x2,
            ey: y2 + n_y2
        });
        a.push({
            c1x: x2 + n_x2 - vx21,
            c1y: y2 + n_y2 - vy21,
            c2x: x2 - n_x2 - vx21,
            c2y: y2 - n_y2 - vy21,
            ex: x2 - n_x2,
            ey: y2 - n_y2
        });
        a.push({
            c1x: x1 - n_x2,
            c1y: y1 - n_y2,
            c2x: x1 - n_x0,
            c2y: y1 - n_y0,
            ex: x0 - n_x0,
            ey: y0 - n_y0
        });
        a.push({
            c1x: x0 - n_x0 - vx01,
            c1y: y0 - n_y0 - vy01,
            c2x: x0 + n_x0 - vx01,
            c2y: y0 + n_y0 - vy01,
            ex: x0 + n_x0,
            ey: y0 + n_y0
        });
        a[0].mx = a[0].mx.toFixed(1);
        a[0].mx = parseFloat(a[0].mx);
        a[0].my = a[0].my.toFixed(1);
        a[0].my = parseFloat(a[0].my);
        for (let i = 1; i < a.length; i++) {
            a[i].c1x = a[i].c1x.toFixed(1);
            a[i].c1x = parseFloat(a[i].c1x);
            a[i].c1y = a[i].c1y.toFixed(1);
            a[i].c1y = parseFloat(a[i].c1y);
            a[i].c2x = a[i].c2x.toFixed(1);
            a[i].c2x = parseFloat(a[i].c2x);
            a[i].c2y = a[i].c2y.toFixed(1);
            a[i].c2y = parseFloat(a[i].c2y);
            a[i].ex = a[i].ex.toFixed(1);
            a[i].ex = parseFloat(a[i].ex);
            a[i].ey = a[i].ey.toFixed(1);
            a[i].ey = parseFloat(a[i].ey);
        }
        return a;
    }
    const bethelDraw = (point, is_fill, color = undefined) => {
        const ctx = ctxx.current!
        ctx.beginPath();
        ctx.moveTo(point[0].mx, point[0].my);
        if (undefined != color) {
            ctx.setFillStyle(color);
            ctx.setStrokeStyle(color);
        } else {
            ctx.setFillStyle(point[0].color);
            ctx.setStrokeStyle(point[0].color);
        }
        for (let i = 1; i < point.length; i++) {
            ctx.bezierCurveTo(point[i].c1x, point[i].c1y, point[i].c2x, point[i].c2y, point[i].ex, point[i].ey);
        }
        ctx.stroke();
        if (undefined != is_fill) {
            ctx.fill(); //填充图形 ( 后绘制的图形会覆盖前面的图形, 绘制时注意先后顺序 )
        }
        ctx.draw(true);
    }

    //完成
    const subCanvas = () => {
        return new Promise((resolve, reject) => {
            if (linePrack.current.length === 0) {
                if (defaultSign) {
                    resolve(defaultSign)
                } else {
                    Taro.showToast({ title: '请签名', icon: 'none', mask: true })
                    reject('')
                }
                return
            }
            Taro.canvasToTempFilePath({
                canvasId: canvasName,
                fileType: 'png',
                quality: 1,
                success(res) {
                    Taro.showLoading({ title: '正在上传' })
                    Taro.uploadFile({
                        url: api.uploadUrl,
                        name: "file",
                        filePath: res.tempFilePath,
                        success: (res) => {
                            const resData = JSON.parse(res.data);
                            if (resData.code === 200) {
                                resolve(resData.data)
                            }
                        },
                        complete: () => {
                            Taro.hideLoading()
                        }
                    })
                },
                fail(err) {
                    reject(err)
                }
            })
        })
    }

    //预览图片
    const previewSign = () => {
        if (linePrack.current.length === 0) {
            if (defaultSign) {
                Taro.canvasToTempFilePath({
                    canvasId: canvasName,
                    fileType: 'png',
                    quality: 1,
                    success(res) {
                        Taro.previewImage({ urls: [res.tempFilePath] })
                    }
                })
            } else {
                Taro.showToast({ title: '请签名', icon: 'none', mask: true })
            }
            return
        }
        Taro.canvasToTempFilePath({
            canvasId: canvasName,
            fileType: 'png',
            quality: 1,
            success(res) {
                Taro.previewImage({ urls: [res.tempFilePath] })
            }
        });
    }

    //重写清除画布
    const resetDraw = () => {
        linePrack.current = []
        ctxx.current?.clearRect(0, 0, canvasWidth.current, canvasHeight.current);
        ctxx.current?.rect(0, 0, canvasWidth.current, canvasHeight.current);
        ctxx.current?.setFillStyle(bgColor);
        ctxx.current?.fill(); //设置填充
        ctxx.current?.draw();
    }

    useImperativeHandle(ref, () => {
        return {
            subCanvas,
            previewSign,
            resetDraw
        }
    })

    return <Canvas
        // width={width}
        // height={height}
        // type='2d'
        style={{ width, height }}
        className={canvasName}
        canvas-id={canvasName}
        disable-scroll
        onTouchStart={uploadScaleStart}
        onTouchMove={uploadScaleMove}
        onTouchEnd={uploadScaleEnd}
    />
})

export default ElectronicSignature